from dataclasses import field, dataclass
from typing import Dict, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from tofuref.data.providers import Provider


@dataclass
class Registry:
    """Not just for providers, so kinda also application state"""

    fullscreen_mode: bool = False
    providers: Dict[str, "Provider"] = field(default_factory=dict)
    active_provider: Optional["Provider"] = None


registry = Registry()



from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Optional

from tofuref.data.helpers import get_registry_api, header_markdown_split

if TYPE_CHECKING:
    from tofuref.data.providers import Provider


class ResourceType(Enum):
    RESOURCE = "resource"
    DATASOURCE = "datasource"
    GUIDE = "guide"
    FUNCTION = "function"


@dataclass
class Resource:
    name: str
    provider: "Provider"
    type: ResourceType
    _content: Optional[str] = None

    def __lt__(self, other: "Resource") -> bool:
        return self.name < other.name

    def __gt__(self, other: "Resource") -> bool:
        return self.name > other.name

    def __str__(self):
        return f"[cyan]{self.type.value[0].upper()}[/] {self.name}"

    def __rich__(self):
        return str(self)

    def __hash__(self):
        return hash(f"{self.provider.name}_{self.type}_{self.name}")

    async def content(self):
        if self._content is None:
            doc_data = await get_registry_api(
                f"{self.provider.organization}/{self.provider.name}/{self.provider.active_version}/{self.type.value}s/{self.name}.md",
                json=False,
            )
            _, self._content = header_markdown_split(doc_data)
        return self._content



from textual.binding import Binding
from textual.widgets import (
    Input,
    OptionList,
    RichLog,
    MarkdownViewer,
)

from tofuref import __version__

log_widget = RichLog(id="log", markup=True, wrap=True, classes="bordered hidden")
log_widget.border_title = "Log"
log_widget.border_subtitle = f"tofuref v{__version__}"
log_widget.display = False


class CustomMarkdownViewer(MarkdownViewer):
    ALLOW_MAXIMIZE = True

    BINDINGS = [
        Binding("up", "up", "Scroll Up", show=False),
        Binding("down", "down", "Scroll Down", show=False),
        Binding("pageup", "page_up", "Page Up", show=False),
        Binding("pagedown", "page_down", "Page Down", show=False),
        Binding("home", "scroll_home", "Top", show=False),
        Binding("end", "scroll_end", "Bottom", show=False),
    ]

    def action_up(self) -> None:
        self.document.scroll_up()

    def action_down(self) -> None:
        self.document.scroll_down()

    def action_page_down(self) -> None:
        self.document.action_page_down()

    def action_page_up(self):
        self.document.action_page_up()

    def action_scroll_home(self) -> None:
        self.document.scroll_home()

    def action_scroll_end(self) -> None:
        self.document.scroll_end()

    # Without this, the Markdown viewer would try to open a file on a disk, while the Markdown itself will open a browser link (desired)
    async def go(self, location):
        return None


content_markdown = CustomMarkdownViewer(
    f"""
# Welcome to tofuref {__version__}!

Changelog: https://github.com/djetelina/tofuref/blob/main/CHANGELOG.md

## Controls
Navigate with arrows/page up/page down/home/end or your mouse.

| keybindings | action |
|------|--------|
| `tab` | focus next window |
| `shift+tab` | focus previous window |
| `enter` | choose selected or finish search |
| `q`, `ctrl+q` | **quit** tofuref |
| `s`, `/` | **search** in the context of providers and resources |
| `v` | change active provider **version** |
| `p` | focus **providers** window |
| `r` | focus **resources** window |
| `c` | focus **content** window |
| `f` | toggle **fullscreen** mode |
| `l` | display **log** window |

---

# Known issues

Contributions (issues, prs, docs) welcome: [GitHub](https://github.com/djetelina/tofuref)

* *[UX]* Navigating content is difficult (search doesn't work despite the footer suggesting otherwise
* *[BUG]* Pressing `<esc>` while searching in fullscreen mode breaks the fullscreen and hides the still focused search

---

# Get in touch
* Matrix: #tofuref:jtl.vision
* GitHub: https://github.com/djetelina/tofuref/issues

## Feedback requests
* tofuref can start in fullscreen mode depending on the width of your terminal window, is the default good?

""",
    classes="content",
    show_table_of_contents=False,
)


search = Input(placeholder="Search...", id="search", classes="bordered")
search.border_title = "Search"

navigation_providers = OptionList(
    name="Providers", id="nav-provider", classes="nav-selector bordered"
)
navigation_providers.border_title = "Providers"
navigation_resources = OptionList(
    name="Resources", id="nav-resources", classes="nav-selector bordered"
)
navigation_resources.border_title = "Resources"



import re
from typing import Tuple, Union, Dict

import httpx
from yaml import safe_load
from yaml.scanner import ScannerError

from tofuref.widgets import log_widget


def header_markdown_split(contents: str) -> Tuple[dict, str]:
    header = {}
    if re.match(r"^---$", contents, re.MULTILINE):
        split_contents = re.split(r"^---$", contents, 3, re.MULTILINE)
        try:
            header = safe_load(split_contents[1])
        except ScannerError as _:
            header = {}
        markdown_content = split_contents[2]
    else:
        markdown_content = contents
    return header, markdown_content


async def get_registry_api(endpoint: str, json=True) -> Union[Dict[str, dict], str]:
    uri = f"https://api.opentofu.org/registry/docs/providers/{endpoint}"
    async with httpx.AsyncClient() as client:
        r = await client.get(uri)

    log_widget.write(f"GET [cyan]{endpoint}[/] [bold]{r.status_code}[/]")
    if json:
        return r.json()
    else:
        return r.text



from dataclasses import dataclass, field
from typing import Dict, Optional, List
from rich.json import JSON as RICH_JSON
import json

from tofuref.data.resources import Resource, ResourceType
from tofuref.data.helpers import (
    get_registry_api,
    header_markdown_split,
)


@dataclass
class Provider:
    organization: str
    name: str
    description: str
    fork_count: int
    blocked: bool
    popularity: int
    _overview: Optional[str] = None
    _active_version: Optional[str] = None
    versions: List[Dict[str, str]] = field(default_factory=list)
    fork_of: Optional[str] = None
    raw_json: Optional[dict] = None
    resources: List[Resource] = field(default_factory=list)
    datasources: List[Resource] = field(default_factory=list)
    functions: List[Resource] = field(default_factory=list)
    guides: List[Resource] = field(default_factory=list)

    @classmethod
    def from_json(cls, data: dict) -> "Provider":
        return cls(
            organization=data["addr"]["namespace"],
            name=data["addr"]["name"],
            description=data["description"],
            fork_count=data["fork_count"],
            blocked=data["is_blocked"],
            popularity=data["popularity"],
            versions=data["versions"],
            fork_of=data.get("fork_of", {}).get("display"),
            raw_json=data,
        )

    @property
    def display_name(self) -> str:
        return f"{self.organization}/{self.name}"

    @property
    def active_version(self):
        if self._active_version is None:
            self._active_version = self.versions[0]["id"]
        return self._active_version

    @active_version.setter
    def active_version(self, value):
        self._active_version = value
        self.resources = []
        self._overview = None

    @property
    def _endpoint(self) -> str:
        return f"{self.organization}/{self.name}/{self.active_version}"

    async def overview(self):
        if self._overview is None:
            self._overview = await get_registry_api(
                f"{self._endpoint}/index.md", json=False
            )
            _, self._overview = header_markdown_split(self._overview)
        return self._overview

    async def load_resources(self):
        if self.resources:
            return
        resource_data = await get_registry_api(
            f"{self.organization}/{self.name}/{self.active_version}/index.json"
        )
        for g in sorted(resource_data["docs"]["guides"], key=lambda x: x["name"]):
            self.resources.append(Resource(g["name"], self, type=ResourceType.GUIDE))

        for r in sorted(resource_data["docs"]["resources"], key=lambda x: x["name"]):
            self.resources.append(Resource(r["name"], self, type=ResourceType.RESOURCE))
        for d in sorted(resource_data["docs"]["datasources"], key=lambda x: x["name"]):
            self.resources.append(
                Resource(d["name"], self, type=ResourceType.DATASOURCE)
            )
        for f in sorted(resource_data["docs"]["functions"], key=lambda x: x["name"]):
            self.resources.append(Resource(f["name"], self, type=ResourceType.FUNCTION))

    def __rich__(self):
        return RICH_JSON(
            json.dumps(
                {
                    k: v
                    for k, v in self.__dict__.items()
                    if k not in ["raw_json", "versions"]
                }
            )
        )


async def populate_providers() -> Dict[str, Provider]:
    providers = {}

    data = await get_registry_api("index.json")

    for provider_json in data["providers"]:
        provider = Provider.from_json(provider_json)
        filter_in = (
            provider.versions,
            not provider.blocked,
            (not provider.fork_of or provider.organization == "opentofu"),
            provider.organization not in ["terraform-providers"],
        )
        if all(filter_in):
            providers[provider.display_name] = provider

    providers = {
        k: v
        for k, v in sorted(
            providers.items(), key=lambda p: p[1].popularity, reverse=True
        )
    }

    return providers



import asyncio
from typing import Iterable, Optional, List

import textual.app
from textual import on
from textual.app import App, ComposeResult
from textual.containers import Container
from textual.widgets import (
    Footer,
    Input,
    OptionList,
    Select,
)
from rich.markdown import Markdown

from tofuref.data.resources import Resource
from tofuref.data.providers import populate_providers, Provider
from tofuref.data.registry import registry
from tofuref.widgets import (
    log_widget,
    content_markdown,
    navigation_providers,
    navigation_resources,
    search,
)


class TofuRefApp(App):
    CSS_PATH = "tofuref.tcss"
    TITLE = "TofuRef - OpenTofu Provider Reference"
    BINDINGS = [
        ("q", "quit", "Quit"),
        ("s", "search", "Search"),
        ("/", "search", "Search"),
        ("v", "version", "Provider Version"),
        ("p", "providers", "Providers"),
        ("r", "resources", "Resources"),
        ("c", "content", "Content"),
        ("f", "fullscreen", "Fullscreen Mode"),
        ("l", "log", "Show Log"),
    ]

    def compose(self) -> ComposeResult:
        # Navigation
        with Container(id="sidebar"):
            with Container(id="navigation"):
                yield navigation_providers
                yield navigation_resources

        # Main content area
        with Container(id="content"):
            yield content_markdown

        yield log_widget

        yield Footer()

    async def on_ready(self) -> None:
        log_widget.write("Populating providers from the registry API")
        content_markdown.document.classes = "bordered content"
        content_markdown.document.border_title = "Content"
        content_markdown.document.border_subtitle = "Welcome"
        if self.size.width < 125:
            registry.fullscreen_mode = True
        if registry.fullscreen_mode:
            navigation_providers.styles.column_span = 2
            navigation_resources.styles.column_span = 2
            content_markdown.styles.column_span = 2
            self.screen.maximize(navigation_providers)
        navigation_providers.loading = True
        self.screen.refresh()
        await asyncio.sleep(0.1)
        self.app.run_worker(self._preload, name="preload")

    @staticmethod
    async def _preload():
        registry.providers = await populate_providers()
        log_widget.write(f"Providers loaded ([cyan bold]{len(registry.providers)}[/])")
        _populate_providers()
        navigation_providers.loading = False
        navigation_providers.highlighted = 0
        log_widget.write(Markdown("---"))

    def action_search(self) -> None:
        """Focus the search input."""
        if search.has_parent:
            search.parent.remove_children([search])
        for searchable in [navigation_providers, navigation_resources]:
            if searchable.has_focus:
                search.value = ""
                searchable.mount(search)
                search.focus()
                search.offset = searchable.offset + (
                    0,
                    searchable.size.height - 3,
                )

    def action_log(self) -> None:
        log_widget.display = not log_widget.display

    def action_providers(self):
        if registry.fullscreen_mode:
            self.screen.maximize(navigation_providers)
        navigation_providers.focus()

    def action_resources(self):
        if registry.fullscreen_mode:
            self.screen.maximize(navigation_resources)
        navigation_resources.focus()

    def action_content(self):
        if registry.fullscreen_mode:
            self.screen.maximize(content_markdown)
        content_markdown.document.focus()

    def action_fullscreen(self):
        if registry.fullscreen_mode:
            registry.fullscreen_mode = False
            navigation_providers.styles.column_span = 1
            navigation_resources.styles.column_span = 1
            content_markdown.styles.column_span = 1
            self.screen.minimize()
        else:
            registry.fullscreen_mode = True
            navigation_providers.styles.column_span = 2
            navigation_resources.styles.column_span = 2
            content_markdown.styles.column_span = 2
            self.screen.maximize(self.screen.focused)

    async def action_version(self):
        if registry.active_provider is None:
            self.notify(
                "Provider Version can only be changed after one is selected.",
                title="No provider selected",
                severity="warning",
            )
            return
        if navigation_resources.children:
            navigation_resources.remove_children("#version-select")
        else:
            version_select = Select.from_values(
                (v["id"] for v in registry.active_provider.versions),
                prompt="Select Provider Version",
                allow_blank=False,
                value=registry.active_provider.active_version,
                id="version-select",
            )
            navigation_resources.mount(version_select)
            await asyncio.sleep(0.1)
            version_select.action_show_overlay()

    @on(Select.Changed, "#version-select")
    async def change_provider_version(self, event: Select.Changed) -> None:
        if event.value != registry.active_provider.active_version:
            registry.active_provider.active_version = event.value
            await _load_provider_resources(registry.active_provider)
            navigation_resources.remove_children("#version-select")

    def on_input_changed(self, event: Input.Changed) -> None:
        if event.input.id != "search":
            return

        query = event.value.strip()
        if search.parent == navigation_providers:
            if not query:
                _populate_providers()
            else:
                _populate_providers(
                    [p for p in registry.providers.keys() if query in p]
                )
        elif search.parent == navigation_resources:
            if not query:
                _populate_resources(registry.active_provider)
            else:
                _populate_resources(
                    registry.active_provider,
                    [r for r in registry.active_provider.resources if query in r.name],
                )

    def on_input_submitted(self, event: Input.Submitted) -> None:
        search.parent.focus()
        search.parent.highlighted = 0
        search.parent.remove_children([search])

    async def on_option_list_option_selected(
        self, event: OptionList.OptionSelected
    ) -> None:
        if event.control == navigation_providers:
            provider_selected = registry.providers[event.option.prompt]
            registry.active_provider = provider_selected
            if registry.fullscreen_mode:
                self.screen.maximize(navigation_resources)
            await _load_provider_resources(provider_selected)
        elif event.control == navigation_resources:
            resource_selected = event.option.prompt
            if registry.fullscreen_mode:
                self.screen.maximize(content_markdown)
            content_markdown.loading = True
            content_markdown.document.update(await resource_selected.content())
            content_markdown.document.border_subtitle = f"{resource_selected.type.value} - {resource_selected.provider.name}_{resource_selected.name}"
            content_markdown.document.focus()
            content_markdown.loading = False


async def _load_provider_resources(provider: Provider):
    navigation_resources.loading = True
    content_markdown.loading = True
    await provider.load_resources()
    content_markdown.document.update(await provider.overview())
    content_markdown.document.border_subtitle = (
        f"{provider.display_name} {provider.active_version} Overview"
    )
    _populate_resources(provider)
    navigation_resources.focus()
    navigation_resources.highlighted = 0
    content_markdown.loading = False
    navigation_resources.loading = False


def _populate_providers(providers: Optional[Iterable[str]] = None) -> None:
    if providers is None:
        providers = registry.providers.keys()
    navigation_providers.clear_options()
    navigation_providers.border_subtitle = f"{len(providers)}/{len(registry.providers)}"
    for name in providers:
        navigation_providers.add_option(name)


def _populate_resources(
    provider: Optional[Provider] = None, resources: Optional[List[Resource]] = None
) -> None:
    navigation_resources.clear_options()
    if provider is None:
        return
    navigation_resources.border_subtitle = (
        f"{provider.organization}/{provider.name} {provider.active_version}"
    )

    if resources is None:
        for resource in provider.resources:
            navigation_resources.add_option(resource)
    else:
        navigation_resources.add_options(resources)


def main():
    TofuRefApp().run()


if __name__ == "__main__":
    main()
